/* $Workfile: TecRecPlay.c $       $Revision: 3 $      $Modtime: 9/17/08 12:18p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    TecRecPlay.c

    A simple example console application demonstrating the use of the
    DMA Direct Sending and Receiving with the OTX Driver using the
    board's Burst device to send and receive data to/from all timeslots (TS1-TS31) of
	Li0 and Li1.

    This application is designed for the following board:
      Thor-2-ExpressCard 

	The transmitted data (TS1-TS31) is read from files (which file names are supplied
	as command line parameters). 

    The received data is stored in files (which file names are supplied
	as command line parameters). 


*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxBrd.h"
#include "OtxT1E1.h"  
#include "OtxDsp.h"  

#include "OtxBurst.h"
#include "OtxBurstD.h"



#define DEMO_DEFAULT_LOG_FILE   "TecRecPlay.log"
#define DEMO_DEFAULT_BOARD_NO   0
#define DEMO_DEFAULT_RX_THRESHOLD 500     // in mV

#define DEMO_BURST_BUF_SIZE     0x8000    // 32k, blocks of 128 bytes, where each Li uses 32 of the 128 bytes in the block


typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  char                   liRxFileName[2][OTX_MAX_PATH];
  char                   liTxFileName[2][OTX_MAX_PATH];
  OtxPhysicalDeviceTypeE nBoardType;
  OtxT1E1LiModeE         eT1E1Mode;    // T1 or E1 mode
  int                    coLi;
  int                    coHighways;
  int                    nHwyTimeslots;
  int                    nCurrentHwy;
  int                    boardNo;
  OTX_BOOL               bReceiveEqualizer;
  OtxT1E1TermModeE       eTermMode;
  OTX_BOOL               bMonitorMode;
  OTX_BOOL               bAmplifier;
  OTX_INT32              nRxThreshold;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
  OtxBrdClockSourceE     eClkSource;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoBurstSender(OTX_UINT32 nSourceId);
static void DemoBurstReceiver(OTX_UINT32 nSourceId);
static void DemoSpinner(void);


/*******************************************************************************
 Global Data
*******************************************************************************/

OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
int         g_cbMaxTxFileSize = 0;


struct {
  OTX_HANDLE hThorController; // This handle represents the board itself. It hosts devices like the Burst device and the Dual T1E1 device
  OTX_HANDLE hBurst;          // There is one burst device on this board which handles both transmit (hostPC->Thor-2ExpressCard) and receive (Thor-2ExpressCard->hostPC)
  OTX_HANDLE hLi[2];          // There are 2 Line Interfaces (Li) on the Thor-2-ExpressCard
  OTX_HANDLE hDualE1T1;       // The two Li are hosted by a Dual T1E1 device
} g_hPhysDevs;

// Burst DMA Sender
struct sBurstSender {
  OTX_UINT32  nSourceId;  // Always zero on this board since this board only has one burst highway
  OTX_BOOL    bVerbose;   // Not used
  OTX_BOOL    bActive;
  OTX_UINT8   sqTxBuf[DEMO_BURST_BUF_SIZE];
  FILE        *pTxFile[2];  // File pointer to the Transmit File. One file for each Li
  OTX_INT32   cbTxFile[2];  // File size of the Transmit File. One file for each Li
  OTX_BOOL    bFileLoop;    // Set to OTX_TRUE if the sender should rewind the file at EOF and continue to send continuously

};

// Burst DMA Receiver
struct sBurstReceiver {
  OTX_UINT32  nSourceId;   // Always zero on this board since this board only has one burst highway
  OTX_BOOL    bVerbose;    // Not used, only for debugging
  OTX_BOOL    bActive;
  FILE        *pRxFile[2]; // File pointer to the Receive File. One file for each Li
};

struct {
  struct sBurstSender mBurstSender;      
  struct sBurstReceiver mBurstReceiver;  
} g_hLogicDevs;

// This struct is used to configure the T1 and E1 Layer 1 parameters of the two Line interfaces
struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************
*******************************************************************************/



static void DemoCheckRc(IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from en API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {
    OTX_RESULT otxRc;
    char msgStr[256];
    
    // Convert the result code to a more descriptive string
    otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
    if (OTX_FAILURE(otxRc)) {
      strcpy(msgStr, "Cannot translate this result code.");
    }
    
    DemoLogf("ERROR(%#lX [%d.%d]): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), msgStr);
    
    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;

  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000LL;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;
  #endif

  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);

  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);

  if (ptCurrentDayAndTime) {

    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ",
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min,
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }

}


/*******************************************************************************
  @func
    Idle Function that checks for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);

    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(result);
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event

        // Handle Burst Send and Receive Events here
        switch (eventData.m_eDeviceType) {
          case OTX_DEVICE_BURST:
            switch (eventData.m_nCode) {
            case OTX_BURST_RECEIVER_EC_DATA_AVAILABLE:
              DemoBurstReceiver(eventData.m_nRequestId);
              continue;
            case OTX_BURST_SENDER_EC_BUF_DEPLETED:
              DemoBurstSender(eventData.m_nRequestId);
              continue;
            default:
              break;
            }
            break;
          default:
            break;
        }

        // Handle other events here (E.g. T1/E1 Layer 1 status events)
        DemoPrintEvent(&eventData);
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);

  }

  return 0;
}


/*******************************************************************************
  @func
    Burst Sender Function

    This function is called (via an OTX Drive event) when the the Burst device 
    can accept more data.

    It reads data from one or two files (one per Li), stuffs the data into
    the proper locations in a buffer, and send the buffer to the Burst device.
    The burst device will then send this buffer (via burst DMA) to the Thor-2-ExpressCard.
    The card sends the data to the T1/E1 devices which transmit the Data on each
    respective T1/E1 span.
    
*******************************************************************************/
void DemoBurstSender(OTX_UINT32 nSourceId)
{
  OTX_RESULT  otxRc;
  OTX_INT32  nLi;
  int cbTotalLiBuf;

  // validate parameters
  if (nSourceId != 0) {
    DemoLogf("DemoBurstSender: Invalid nSourceId(%d)!\n", nSourceId);
    return;
  }

  if (!g_hLogicDevs.mBurstSender.bActive) {
    // we're not active, so just return
    return;
  }

START:
  // Prime the transmit buffer with 0xFF (in case we are not sending from both Lis)
  memset(&g_hLogicDevs.mBurstSender.sqTxBuf[0], 0xFF, sizeof(g_hLogicDevs.mBurstSender.sqTxBuf) / sizeof(g_hLogicDevs.mBurstSender.sqTxBuf[0]));

  cbTotalLiBuf = 0;

  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    if (g_hLogicDevs.mBurstSender.pTxFile[nLi]) {
      OTX_UINT8 singleLiBuf[DEMO_BURST_BUF_SIZE / 4]; // Divide by 4 since the burst highway has 128 timeslot. Each Li occupies only 32 timeslots of this highway
      int cbSingleLiBuf;
      int cbMaxSingleLiBuf = DEMO_BURST_BUF_SIZE / 4;
      int byteCount;
      int x;
      
      // Read a number of 32-octet blocks from the file representing all timeslots that will be transmitted 
      // from a specified Li.
      // 
      // E1 mode:
      // ========
      // Please note that we data for TS0 (index zero of the 32-octet block), and we put it in the burst buffer, 
      // but it is actually not transmitted (unless the T1E1 device is configured for transparent framing mode).
      // 
      // T1 mode:
      // ========
      // If we are running in T1 mode we still read 32-timeslot blocks from the file. The entire 32-timeslot block
      // is placed in the burst buffer. The normal T1 timeslots are located in TS1 through TS24 of this 32-octet 
      // block.
      // Index zero of the 32-octet block represents the the DL-bit (as the MSBit of the octet) and it is 
      // only be transitted if the T1 device is configured in transparent mode. Index 25-31 are never transmitted
      // in T1 mode, but the octets are still placed in the burst buffer.
      //
      cbSingleLiBuf = fread(
        &singleLiBuf,
        sizeof(singleLiBuf[0]),
        cbMaxSingleLiBuf,
        g_hLogicDevs.mBurstSender.pTxFile[nLi]
        );
      
      if (cbSingleLiBuf > 0) {
        if ((cbSingleLiBuf > 0) && (cbSingleLiBuf < cbMaxSingleLiBuf)) {
          // We are at the end of the file, and we don't have a full buffer to send.
          // Pad with 0xFF at the end of the buffer to make sure that we don't underrun the Burst Sender
          DemoLogf("Pad with 0xFF [%d] at the end of the buffer to make sure that we don't underrun the Burst Sender\n", (cbMaxSingleLiBuf - cbSingleLiBuf) / sizeof(singleLiBuf[0]));
          memset(&singleLiBuf[cbSingleLiBuf], 0xFF, (cbMaxSingleLiBuf - cbSingleLiBuf) / sizeof(singleLiBuf[0]) );
          cbSingleLiBuf = cbMaxSingleLiBuf;
        }
        
        // Dispurse the single Li timeslots (32) to be transmitted in the whole transmit buffer (128 timeslots)
        // The Li0 timeslots on this card are the TS0, TS4, TS8, ... , TS124 timeslots of the burst highway.
        // The Li1 timeslots on this card are the TS2, TS6, TS10, ... , TS126 timeslots of the burst highway.    
        byteCount = 0;
        //DemoLogf("g_demoEnv.coHighways=%d, g_hLogicDevs.mBurstSender.nLiToSend=%d \n", g_demoEnv.nCurrentHwy, g_demoEnv.coLi, g_demoEnv.coHighways, g_hLogicDevs.mBurstSender.nLiToSend);
        for (x = nLi*2; (byteCount < cbSingleLiBuf) && (x < DEMO_BURST_BUF_SIZE); x += 4) {
          g_hLogicDevs.mBurstSender.sqTxBuf[x] = singleLiBuf[byteCount++];
        }
        
        assert(byteCount == cbSingleLiBuf);
        
      }

      cbTotalLiBuf += cbSingleLiBuf;
    }
  }


  if (cbTotalLiBuf > 0) {
    OTX_UINT32  ui32 = DEMO_BURST_BUF_SIZE;
    
    otxRc = OtxBurstSendData(g_hPhysDevs.hBurst, nSourceId, g_hLogicDevs.mBurstSender.sqTxBuf, &ui32);
    DemoCheckRc(otxRc);
    
    // Check the amount buffered w/ how much we requested to be buffered.
    // (since we set the trigger amount to the buffer size, there should
    // have been room to buffer this, so these two should be the same).
    if (ui32 != DEMO_BURST_BUF_SIZE) {
      DemoLogf("DemoBurstSender: ui32 (%d) != DEMO_BURST_BUF_SIZE (%d)\n", ui32, DEMO_BURST_BUF_SIZE);
    }
    
  } else {
    
    if (g_hLogicDevs.mBurstSender.bFileLoop) {
      if (g_hLogicDevs.mBurstSender.bVerbose) {
        DemoLogf("\nDemoBurstSender: Transmission complete. Starting file from the beginning.\n");
      }
      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        if (g_hLogicDevs.mBurstSender.pTxFile[nLi]) {
          rewind(g_hLogicDevs.mBurstSender.pTxFile[nLi]);
        }
      }
      goto START;
    } else {
      DemoLogf("\nDemoBurstSender: Transmission complete. Closing file\n");
      for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
        if (g_hLogicDevs.mBurstSender.pTxFile[nLi]) {
          fclose(g_hLogicDevs.mBurstSender.pTxFile[nLi]);
          g_hLogicDevs.mBurstSender.pTxFile[nLi] = NULL;
        }
      }
      g_hLogicDevs.mBurstSender.bActive = OTX_FALSE;
    }
  }
    
  return;
}

/*******************************************************************************
  @func
    Burst Receiver Function

    The function is called (via an OTX Driver event) when the the Burst device
    has data for us to read. This function reads the data from the burst device
    and then parses the data buffer, and extracts data associated with each Line
    Interface, and writes the data to one or two files (one per Line interface).

    The burst data arrives at a rate of 8192 kbit/s. We call that a burst highway
    with a 8192 kbit/s speed. The highway is formatted into 128-octet blocks, which 
    consequently arrives every 125 microseconds, which is the frame period in E1 
    as well as T1.

    In the 128-octet block (frame) you will find the 32 timeslots (frame) from Li0, 
    and the 32 timeslots (frame) from Li1. The timeslot data is interspersed among
    the 128 octets in the following way:

        Li0 timeslots: TS0, TS4, TS8, ... , TS124 
        Li1 timeslots: TS2, TS6, TS10, ... , TS126     

*******************************************************************************/
void DemoBurstReceiver(OTX_UINT32 nSourceId)
{
  OTX_RESULT    otxRc;
  OTX_UINT8     buf[DEMO_BURST_BUF_SIZE];
  OTX_UINT32    ui32;
  OTX_UINT64    nFrameRefId;
  OTX_DATETIME  dtFrameRefTime;
  OTX_INT32  nLi;
  
  // validate parameters
  if (nSourceId != 0) {
    DemoLogf("DemoBurstReceiver: Invalid nSourceId(%d)...\n", nSourceId);
    return;
  }
  
  if (!g_hLogicDevs.mBurstReceiver.bActive) {
    // we're not active, so just return
    return;
  }
  
  // we want the frame reference id and time
  ui32 = sizeof(buf);
  otxRc = OtxBurstReadDataEx(g_hPhysDevs.hBurst, nSourceId, buf, &ui32, &nFrameRefId, &dtFrameRefTime);
  DemoCheckRc(otxRc);
  
  // check the amount buffered w/ how much we requested to be buffered
  //  (since we set the trigger amount to the buffer size, there should
  //    have been room to buffer this, so these two should be the same).
  if (ui32 != sizeof(buf)) {
    DemoLogf("Burst Receiver: Warning. Received a shorter buffer than expected (buflen=%d, expected=%d)\n", ui32, sizeof(buf));
  }
  if ((ui32 % g_demoEnv.nHwyTimeslots) != 0) {
    DemoLogf("Burst Receiver: Error. The received buffer length is not a multiple of the highway speed (g_demoEnv.nHwyTimeslots TSs) (buflen=%d, expected=%d)\n", g_demoEnv.nHwyTimeslots, ui32, sizeof(buf));
    DemoExitQuestion();
  }
  
  // Echo if we're in verbose mode
  if (g_hLogicDevs.mBurstReceiver.bVerbose) {
    int x;
    
    DemoPrintTimeStamp(dtFrameRefTime);
    
    // Note: we only display the least significant 32 bits of the frame reference id.
    DemoLogf("Burst Receiver:(buflen=%.5d) ID:%08x ", ui32, (OTX_UINT32) (nFrameRefId & 0xffffffff));
    
    for (x = 0; (x < g_demoEnv.nHwyTimeslots) && (x < (OTX_INT32)ui32); x++) {
      DemoLogf("%.2X ", buf[x]);
    }
    
    DemoLogf("\n");
  }
  
  
  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    // write data to a file if desired
    if (g_hLogicDevs.mBurstReceiver.pRxFile[nLi]) {
      int cbWritten;
      
      // Single TS
      OTX_UINT8 singleLiBuf[DEMO_BURST_BUF_SIZE/4];
      int byteCount = 0;
      int x;
      
      byteCount = 0;
      for (x = nLi*2; x < (OTX_INT32)ui32; x += 4) {
        singleLiBuf[byteCount++] = buf[x];
      }
      cbWritten = fwrite(singleLiBuf, sizeof(singleLiBuf[0]), byteCount, g_hLogicDevs.mBurstReceiver.pRxFile[nLi]);
      if (cbWritten != byteCount) {
        DemoLogf("Burst Receiver Li[%d] Error: Could not write data to file.  (byteCount=%d, cbWritten=%d) \n", nLi, byteCount, cbWritten);
      } else {
        //DemoLogf("Burst Receiver Li[%d] (byteCount=%d, cbWritten=%d) \n", nLi, byteCount, cbWritten);
      }
    }
  }
}

void DemoPrintEvent(
/*******************************************************************************
  @func
    Prints a Driver Event.
*******************************************************************************/
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_LI_T1E1:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    // Event from a T1E1 device
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_BURST:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);

    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("Burst[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
    } else {
      DemoLogf("Burst[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  default:
    // Unhandled event
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }
}




/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and one or more characters.

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP)
{
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_2_EXPRESSCARD;
  envP->coLi              = 2;
  envP->coHighways        = 1;
  envP->nHwyTimeslots     = 128;

  envP->eClkSource        = OTX_CLOCK_SOURCE_INTERNAL;
  envP->bReceiveEqualizer = OTX_FALSE;
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;
  envP->eTermMode         = OTX_T1E1_TERM_100_OHM;
  envP->bMonitorMode      = OTX_FALSE;
  envP->bAmplifier        = OTX_FALSE;
  envP->nRxThreshold      = DEMO_DEFAULT_RX_THRESHOLD; // in mV
  envP->liRxFileName[0][0] = '\0';
  envP->liRxFileName[1][0] = '\0';
  envP->liTxFileName[0][0] = '\0';
  envP->liTxFileName[1][0] = '\0';

  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'P') {   // Loop files (rewind at EOF)
          g_hLogicDevs.mBurstSender.bFileLoop = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == '0') {   // Li0 files
          if (*(argv[i]+2) == 'T') {   // Transmit file
            strcpy(envP->liTxFileName[0], argv[i] + 3);
          } else if (*(argv[i]+2) == 'R') { // Receive file  
            strcpy(envP->liRxFileName[0], argv[i] + 3);
          } else {
            DemoLogf("Error: Invalid command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == '1') {   // Li1 files
          if (*(argv[i]+2) == 'T') {   // Transmit file
            strcpy(envP->liTxFileName[1], argv[i] + 3);
          } else if (*(argv[i]+2) == 'R') {  // Receive file 
            strcpy(envP->liRxFileName[1], argv[i] + 3);
          } else {
            DemoLogf("Error: Invalid command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i] + 1) == 'A') {
          // Use Front End Amplifier = ON
          envP->bAmplifier = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'E') {
          // Use ReceiveEqualizer  = ON
          envP->bReceiveEqualizer = OTX_TRUE;
        } else if (*(argv[i] + 1) == '2') {
          // Use MonitorMode  = ON
          envP->bMonitorMode = OTX_TRUE;
        } else if (*(argv[i] + 1) == 'H') {
          // RxThreshold (in mV)
          envP->nRxThreshold = strtoul(argv[i] + 2, NULL, 0);
        } else if (*(argv[i] + 1) == 'Y') {
          // TermMode
          if ( *(argv[i]+2) == '1') {
            envP->eTermMode = OTX_T1E1_TERM_100_OHM;
          } else if ( *(argv[i]+2) == '7') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM;
          } else if ( *(argv[i]+2) == 'Z') {
            envP->eTermMode = OTX_T1E1_TERM_HIGHZ;
          } else if ( *(argv[i]+2) == 'S') {
            envP->eTermMode = OTX_T1E1_TERM_75_OHM_BALANCED;
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'O') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else if (*(argv[i]+2) == '1') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
          } else if (*(argv[i]+2) == '2') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_2;
          } else if (*(argv[i]+2) == '3') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_3;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'N') {
            envP->eT1FrameFormat = OTX_T1_FF_UNFRAMED; // Transparent mode
          } else if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'N') {
            envP->eE1FrameFormat = OTX_E1_FF_UNFRAMED; // Transparent mode
          } else if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == 'C') {
            envP->nBoardType = OTX_DEVICE_THOR_2_EXPRESSCARD;
            envP->coHighways = 1;
            envP->nHwyTimeslots = 128;
          }
          else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
          exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
        exit(1);
      }
    }
  }
}


/*******************************************************************************

  @func

  Displays help for the command line switches.

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[C] : Board Type used for the demo, where\n");
  DemoLogf("                  C = Use Thor-2-ExpressCard Board\n");
  //  DemoLogf("                  R = Use Thor-4-ExpressCard Board\n");
  DemoLogf(" /ME          : E1 Li mode\n");
  DemoLogf(" /MT          : T1 Li mode\n");
  DemoLogf(" /0T<fileName>: Li0 Transmit file\n");
  DemoLogf(" /1T<fileName>: Li1 Transmit file\n");
  DemoLogf(" /0R<fileName>: Li0 Receive file\n");
  DemoLogf(" /1R<fileName>: Li1 Receive file\n");
  DemoLogf(" /1R<fileName>: Li1 Receive file\n");
  DemoLogf(" /P           : Rewind Transmit File at EOF (longest file)\n");
  DemoLogf(" /O[x]        : Clock Source, where\n");
  DemoLogf("                  x = I Internal clock source (default) or\n");
  DemoLogf("                  x = 0 through 1 indicating with Li to recover the clock from\n");
  DemoLogf(" /A           : Front End Amplifier = ON (default is OFF)\n");
  DemoLogf(" /E           : Receive Equalizer = ON (longhaul) (default is OFF)\n");
  DemoLogf(" /2           : -20dB MonitorMode = ON (default is OFF)\n");
  DemoLogf(" /H           : Receive Input Threshold in mV (default is %d mV)\n", DEMO_DEFAULT_RX_THRESHOLD);
  DemoLogf(" /Y           : Termination Mode, where\n");
  DemoLogf("                  1 = 100/120 ohm balanced (default)\n");
  DemoLogf("                  7 = 75 ohm unbalanced\n");
  DemoLogf("                  Z = High-impedance\n");
  DemoLogf("                  S = 75 ohm balanced\n");
  DemoLogf(" /F[N|E|S|9]  : T1 Frame Format, where\n");
  DemoLogf("                  N = Not framed/Unframed\n");
  DemoLogf("                  E = ESF (default)\n");
  DemoLogf("                  4 = F4\n");
  DemoLogf("                  S = SF [D3/D4] (F12)\n");
  DemoLogf("                  9 = SLC96 (F72)\n");
  DemoLogf(" /J[A|B]      : T1 Line Code, where\n");
  DemoLogf("                  A = AMI\n");
  DemoLogf("                  B = B8ZS (default)\n");
  DemoLogf(" /G[N|D|C|G]  : E1 Frame Format, where\n");
  DemoLogf("                  N = Not framed/Unframed\n");
  DemoLogf("                  D = Doubleframe\n");
  DemoLogf("                  C = CRC4 Multiframe (default)\n");
  DemoLogf("                  G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]      : E1 Line Code, where\n");
  DemoLogf("                  A = AMI\n");
  DemoLogf("                  H = HDB3 (default)\n");
  DemoLogf(" /Q           : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  if (g_pDemoLogFile) {
    vfprintf(g_pDemoLogFile,   formatStr, argList);

    fflush(g_pDemoLogFile);
  }
  va_end(argList);

}


int main(int argc, char *argv[])
/*******************************************************************************
*******************************************************************************/
{
  OTX_RESULT result;  // OTX Standard Return Code
  OTX_INT32 nLi;
  OTX_INT32 nTxFiles = 0;  // Number of files that we are transmitting (0, 1, or 2). Max 2 since this board has 2 LIs.
  OTX_INT32 nRxFiles = 0;  // Number of files that we are receiving (0, 1, or 2). Max 2 since this board has 2 LIs.

  OsaConsoleInit();

  printf("TecRecPlay, %s, Odin TeleSystems Inc.\n", __DATE__);
  printf("Example program of T1/E1 Data Sending/Receiving via the Burst device.\n");

  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(result);

    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib();
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Create EventQueue\n");
  {
	  
	  // Create a notification Event
	  result = OtxDrvCreateEvent(&g_hNotificationEvent);
	  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
		  g_hNotificationEvent = NULL;
		  result = OTX_S_OK;
	  }
	  
	  DemoCheckRc(result);
	  
	  // Create Event Queue for driver notifications
	  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(result);
	  
	  // Enable Event Queue
	  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
	  
  }

  // Connect (get handles) to the physical devices on the Thor-2-ExpressCard
  DemoLogf("Open, configure, and enable Physical Devices\n");
  {
    // Loop the burst data in the T1E1 device device (62 of the 128 timeslots)
    OTX_INT32 i;
    
    // Connect to the Physical Board Controller
    switch (g_demoEnv.nBoardType) {
    case OTX_DEVICE_THOR_2_EXPRESSCARD:
      DemoLogf("Opening Thor-2-ExpressCard board\n");
      break;
    default:
      DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
      OsaConsoleExit();
      exit(1);
      break;
    }
    
    result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController); DemoCheckRc(result);
    
    // Get a handle to the physical Burst device.
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_BURST, 0, 0, g_hEventQueue, &g_hPhysDevs.hBurst); DemoCheckRc(result);
    
    // set device buffer event trigger levels
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_RECEIVER_FIFO_TRIGGER,  DEMO_BURST_BUF_SIZE); DemoCheckRc(result);
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hBurst, OTX_ATTR_BURST_SENDER_FIFO_TRIGGER,    DEMO_BURST_BUF_SIZE); DemoCheckRc(result);
    
    // We'll source the E1/T1 device from the Burst device (as opposed to the DSP device) 
    //  Burst Send  ---> E1T1Device 
    result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hThorController, OTX_ATTR_BRD_FALC_SOURCED, OTX_BRD_FALC_SOURCED_P2S); DemoCheckRc(result);

    // Open E1T1 devices; first the Dual T1E1 device
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController, OTX_DEVICE_DUAL_T1E1, 0, 0, g_hEventQueue, &g_hPhysDevs.hDualE1T1); DemoCheckRc(result);
    result = OtxDrvEnable(g_hPhysDevs.hDualE1T1, OTX_TASK_SYNC, 5000L); DemoCheckRc(result);

    // Then open Li0 and Li1 (which are hosted by the Dual T1E1 device)
    for (i = 0; i < g_demoEnv.coLi; i++) {
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hDualE1T1, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(result);
    }

    // Initialize E1 Configuration Data
    g_T1E1CfgData.e1CfgData.eTxLineCode                = g_demoEnv.eE1LineCode;
    g_T1E1CfgData.e1CfgData.eRxLineCode                = g_demoEnv.eE1LineCode;
    g_T1E1CfgData.e1CfgData.eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
    g_T1E1CfgData.e1CfgData.eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
    g_T1E1CfgData.e1CfgData.eAisDetection              = OTX_E1_AIS_ETS300233;
    g_T1E1CfgData.e1CfgData.bfSiBits                   = 0x3;
    g_T1E1CfgData.e1CfgData.bfSaBits                   = 0x1F;
    g_T1E1CfgData.e1CfgData.bExtendedHDB3errorDetect   = OTX_FALSE;
    g_T1E1CfgData.e1CfgData.bAutoRegainMultiframe      = OTX_FALSE;
    g_T1E1CfgData.e1CfgData.bAutoResync                = OTX_TRUE;
    g_T1E1CfgData.e1CfgData.bAutoCrcBitInversion       = OTX_FALSE;
    g_T1E1CfgData.e1CfgData.bAutoRemoteAlarm           = OTX_FALSE;
    g_T1E1CfgData.e1CfgData.nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
    g_T1E1CfgData.e1CfgData.nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
    g_T1E1CfgData.e1CfgData.nLineLength                = 10;
    g_T1E1CfgData.e1CfgData.bHighTxPower               = OTX_FALSE;
    g_T1E1CfgData.e1CfgData.bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;
    
    // Initialize T1 Configuration Data
    g_T1E1CfgData.t1CfgData.eSignalingMode             = OTX_T1_SM_CCS;
    g_T1E1CfgData.t1CfgData.eTxLineCode                = g_demoEnv.eT1LineCode;
    g_T1E1CfgData.t1CfgData.eRxLineCode                = g_demoEnv.eT1LineCode;
    g_T1E1CfgData.t1CfgData.eFrameFormat               = g_demoEnv.eT1FrameFormat;
    g_T1E1CfgData.t1CfgData.bCRC6Enable                = OTX_TRUE;
    g_T1E1CfgData.t1CfgData.eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
    g_T1E1CfgData.t1CfgData.eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
    g_T1E1CfgData.t1CfgData.bAutoResync                = OTX_TRUE;
    g_T1E1CfgData.t1CfgData.bAutoCrcBitInversion       = OTX_FALSE;
    g_T1E1CfgData.t1CfgData.eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
    g_T1E1CfgData.t1CfgData.bAutoRemoteAlarm           = OTX_FALSE;
    g_T1E1CfgData.t1CfgData.nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
    g_T1E1CfgData.t1CfgData.nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
    g_T1E1CfgData.t1CfgData.nLineLength                = 10;
    g_T1E1CfgData.t1CfgData.bHighTxPower               = OTX_FALSE;
    g_T1E1CfgData.t1CfgData.bReceiveEqualizer          = g_demoEnv.bReceiveEqualizer;
    g_T1E1CfgData.t1CfgData.bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels

    // Configure Li0 and Li1 with the specified E1/T1 Layer1 parameters
    for (i = 0; i < g_demoEnv.coLi; i++) {
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) { 
        // Configure T1 data
        result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(result);
      } else {
        // Configure E1 data
        result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(result);
      
      }
    
      // Set the T1 or E1 mode
      result =  OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(result);
    
      // Set line termination mode
      
      DemoLogf("Li[%d]: Setting Line Termination mode to ", i);
      switch (g_demoEnv.eTermMode) {
      case OTX_T1E1_TERM_120_OHM:
        DemoLogf("OTX_T1E1_TERM_120_OHM\n");
        break;
      case OTX_T1E1_TERM_100_OHM:
        DemoLogf("OTX_T1E1_TERM_100_OHM\n");
        break;
      case OTX_T1E1_TERM_75_OHM:
        DemoLogf("OTX_T1E1_TERM_75_OHM\n");
        break;
      case OTX_T1E1_TERM_HIGHZ:
        DemoLogf("OTX_T1E1_TERM_HIGHZ\n");
        break;
      default:
        DemoLogf("Invalid Termination mode.\n");
        break;
      }
      
      result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], g_demoEnv.eTermMode); DemoCheckRc(result);

      if (g_demoEnv.eTermMode == OTX_T1E1_TERM_HIGHZ) {
        DemoLogf("Li[%d]: Tristating Transmit pins in High-Z Termination mode", i);
        // Tri-state the transmit pins of the Line interface (so it does not disturb the receive pins in High-Z mode)
        result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_FALSE); DemoCheckRc(result);
      } else {
        // Drive the transmit pins of the Line interface in all other termination modes
        result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi[i], OTX_TRUE); DemoCheckRc(result);
      }
          
      // Set receive threshold
      DemoLogf("Li[%d]: Setting RX Input Threshold to %d\n", i, g_demoEnv.nRxThreshold);
      result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, g_demoEnv.nRxThreshold); DemoCheckRc(result);
      { // Read back the threshold
        OTX_UINT32 nRxThreshold;
        result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, &nRxThreshold); DemoCheckRc(result);
        DemoLogf("Li[%d]: RX Input Threshold = %d\n", i, nRxThreshold);
      }
    
      // Set Monitor (-20dB) mode
      DemoLogf("Li[%d]: Setting Monitor mode to %s\n", i, g_demoEnv.bMonitorMode ? "ON" : "OFF");
      result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi[i], g_demoEnv.bMonitorMode); DemoCheckRc(result);
    
      // Turning front end Amplifier ON/OFF      
      DemoLogf("Li[%d]: Setting Front End Amplifier to %s\n", i, g_demoEnv.bAmplifier ? "ON" : "OFF");
      result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi[i], OTX_ATTR_T1E1_RX_AMPLIFIER, g_demoEnv.bAmplifier); DemoCheckRc(result);

    }

    // Enable the board (Physical Device)
    DemoLogf("Enable g_hPhysDevs.hThorController\n");
    result = OtxDrvEnable(g_hPhysDevs.hThorController,  OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    
    DemoLogf("Enable g_hPhysDevs.hBurst\n");
    result = OtxDrvEnable(g_hPhysDevs.hBurst, OTX_TASK_SYNC, 5000L);
    
  }

  // Set clock source for the board
  {
    char szClockSource[64];

    switch (g_demoEnv.eClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      strcpy(szClockSource, "Internal");
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
    case OTX_CLOCK_SOURCE_LOCAL_1:
    case OTX_CLOCK_SOURCE_LOCAL_2:
    case OTX_CLOCK_SOURCE_LOCAL_3:
    case OTX_CLOCK_SOURCE_LOCAL_4:
    case OTX_CLOCK_SOURCE_LOCAL_5:
    case OTX_CLOCK_SOURCE_LOCAL_6:
    case OTX_CLOCK_SOURCE_LOCAL_7:
      sprintf(szClockSource, "Line Interface %d", (int)g_demoEnv.eClkSource - (int)OTX_CLOCK_SOURCE_LOCAL_0);
      break;
    case OTX_CLOCK_SOURCE_A_CLOCKS_ETCF:
      sprintf(szClockSource, "H.100 A-Clocks (ETCF mode)");
      break;
    default:
      strcpy(szClockSource, "Unknown");
      break;
    }

    DemoLogf("Clock Source: %s\n", szClockSource);

    result = OtxBrdSetClocks(g_hPhysDevs.hThorController, OTX_CLOCK_MODE_H100_ETCF, g_demoEnv.eClkSource); DemoCheckRc(result);
  }


  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Connect Devices\n");
  {
    OTX_RESULT result;
    OTX_INT32 i;
    OTX_UINT8 sqTimeSlotMask[32];
    OTX_HANDLE hPipe;
    
    for (i = 0; i < g_demoEnv.coLi; i++) {
      // Although the timeslots connections are hardcoded on this card, we still need to connect
      // to the timeslots of the Li devices that we want to control. If we don't connect any pipes 
      // to the Li devices then it will send an Idle code on each timeslot.
      
      memset(sqTimeSlotMask, 0xFF, sizeof(sqTimeSlotMask));
      
      result = OtxDrvCreatePipe(64*32, &hPipe); DemoCheckRc(result);
      result = OtxDrvConnectPipe(hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[i],  -1, sqTimeSlotMask, 32); DemoCheckRc(result);
      
      result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(result);
    }
    
  }

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  {
    // Read CPLD Build Numbers from the board
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszCpldBuildNumbers;
    
    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(result);
    DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
    free(pszCpldBuildNumbers);
  }

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
  OsaThreadStart(DemoIdleFunction);

  OsaThreadSleep(1000); // Wait for the T1E1 devices to settle

  DemoLogf("\n");

  // Check Layer 1 status
  {
    OtxT1E1Layer1StatusE eLayer1Status;
    OTX_INT32 nLi;
    
    for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
      DemoLogf("Line status of Li #%d is ", nLi);
      result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[nLi], &eLayer1Status); DemoCheckRc(result);
      DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
    }
  }


  // Start Data Receiver
  nRxFiles = 0;
  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    
    if (g_demoEnv.liRxFileName[nLi][0]) {
      if ((g_hLogicDevs.mBurstReceiver.pRxFile[nLi] = fopen(g_demoEnv.liRxFileName[nLi], "wb")) == NULL) {
        DemoLogf("Error: Could not open Li%d Receiver file %s \n", nLi, g_demoEnv.liRxFileName[nLi]);
        DemoExitQuestion();
      } else {
        nRxFiles++;
      }
    }
    
  }

  g_hLogicDevs.mBurstReceiver.bActive = OTX_TRUE;
  if (nRxFiles) {
    DemoLogf("Data Receiver has started (%d files)\n", nRxFiles);
  }

  // Start Data Sender
  nTxFiles = 0;
  for (nLi = 0; nLi < g_demoEnv.coLi; nLi++) {
    
    if (g_demoEnv.liTxFileName[nLi][0]) {
      if ((g_hLogicDevs.mBurstSender.pTxFile[nLi] = fopen(g_demoEnv.liTxFileName[nLi], "rb")) == NULL) {
        DemoLogf("Error: Could not open Li%d Sender file %s \n", nLi, g_demoEnv.liTxFileName[nLi]);
        DemoExitQuestion();
      } else {
        nTxFiles++;

        // Get the file size of this file and save the size of the largest file
        fseek(g_hLogicDevs.mBurstSender.pTxFile[nLi], 0, SEEK_END);
        g_hLogicDevs.mBurstSender.cbTxFile[nLi] = ftell(g_hLogicDevs.mBurstSender.pTxFile[nLi]);
        rewind(g_hLogicDevs.mBurstSender.pTxFile[nLi]);
        if (g_hLogicDevs.mBurstSender.cbTxFile[nLi] > g_cbMaxTxFileSize) {
          g_cbMaxTxFileSize = g_hLogicDevs.mBurstSender.cbTxFile[nLi];
        }
      }
      
    }
  }

  g_hLogicDevs.mBurstSender.bActive = OTX_TRUE;
  g_hLogicDevs.mBurstSender.bVerbose = OTX_TRUE;
  if (nTxFiles) {
    DemoLogf("Data Sender has started (%d files)\n", nRxFiles);
    if (g_hLogicDevs.mBurstSender.bFileLoop) {
      DemoLogf("The Sender Data files start over (rewind) at EOF (every %f seconds)\n", (double)g_cbMaxTxFileSize/256000);
      DemoLogf("Abort program by pressing <CTRL-C>\n");
    } else {
      DemoLogf("The program will exit when EOF is reached (in %f seconds)\n", (double)g_cbMaxTxFileSize/256000);
    }
  } else if (nRxFiles) {
    DemoLogf("Abort the data receiving by pressing <CTRL-C>\n");
  } else {
    DemoLogf("No Sender or Receiver files were supplied as command line parameters.\n");
    DemoExitQuestion();
  }


  while (g_hLogicDevs.mBurstSender.bActive) {
    if (nTxFiles) {
      DemoSpinner();
    }
    OsaThreadSleep(500);
  }

  // Abort threads
  DemoLogf("\n");
  DemoLogf("Exiting...\n");
  {
    OTX_INT32 i;

    OsaThreadSleep(1000);   // Let the receiver run an extra second to make sure that we capture everything we have sent (in the case where we are looped back to ourselves)

    for (i = 0; i < g_demoEnv.coHighways; i++) {
      g_hLogicDevs.mBurstReceiver.bActive = OTX_FALSE;
      g_hLogicDevs.mBurstSender.bActive = OTX_FALSE;
    }

    OsaThreadSleep(1500);   // give threads a chance to exit
  }

  // Clean up
  OsaConsoleExit();
  OtxDrvReleaseLib();

  return 0;
}


static void DemoSpinner(void)
{
  static char spinnerchars[] = { '-', '\\', '|', '/' };
  static int pos = 0;

  putchar(spinnerchars[pos]);
  if (++pos == sizeof(spinnerchars)) {
    pos = 0;
  }
  putchar('\b');
}


