/* $Workfile: T8pLineAutoDetect.c $       $Revision: 14 $      $Modtime: 2/26/09 3:07p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T8PLINEAUTODETECT.C

    A simple example console application demonstrating the use of OTX Driver
    with the Thor-2-PCI, Thor-8-PCI, Thor-2-PCI-Plus, Thor-2-PCI-Express, Thor-8-PCI-Plus, 
    Sleipnir-1-PCI, Gimle-16-PCI-Plus, and Thor-2-PCMCIA boards (STD, CST, PRO, and EX).

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 1 E1 Line and tries several different line settings until 
    the line status is active and no Li status messages occur.
    
    Then it enables 1 HDLC receiver device (on a selected timeslot) and checks 
    that all received frames are good frames (no CRC errors).

    Different verbose levels are available via a command line switch.

    The amount to time spent to check framing and HDLC can be modified
    via command line switches as well.

    If the program is successful in finding a good configuration it will 
    create a separate *.cfg file containing the recommended configuration
    settings. The name of this *.cfg file can be specified via a command line
    switch.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <assert.h>
#include <time.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"
#include "OtxDsp.h"

#define DEMO_DEFAULT_LOG_FILE "T8pLineAutoDetect.log"
#define DEMO_DEFAULT_CFG_FILE "T8pLineAutoDetect.cfg"
#define DEMO_DEFAULT_BOARD_NO 0

#define DEFAULT_HDLC_TEST_TIME 3000 // in milliseconds
#define DEFAULT_FRAMER_TEST_TIME 300 // in milliseconds

#define DEMO_MAX_MULTIPLE_FRAME_READ 128

#define FA_FRS0_REG   0x3c  // Framer Receive Status 0 

#define  FRS0_LOS     0x80
#define  FRS0_AIS     0x40
#define  FRS0_LFA     0x20
#define  FRS0_RRA     0x10
#define  FRS0_NMF     0x04
#define  FRS0_LMFA    0x02


typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL               bExitOnError;
  char                   cwd[OTX_MAX_PATH];
  char                   logFileName[OTX_MAX_PATH];
  char                   cfgFileName[OTX_MAX_PATH];
  OTX_INT32              nBoardType;
  int                    boardNo;
  int                    nTs;
  int                    nTestLi;
  int                    nVerboseLevel;
  int                    nHdlcTestTime;
  int                    nFramerTestTime;
  OtxT1E1LiModeE         eT1E1Mode;    // T1 or E1 mode
  OTX_BOOL               bSs7FisuFiltering;
} DemoEnvT;


typedef struct DemoCfgT {
  OTX_BOOL               bReceiveEqualizer;
  OtxT1E1TermModeE       eTermMode;
  OTX_BOOL               bMonitorMode;
  OTX_BOOL               bAmplifier;
  OTX_INT32              nRxThreshold;


  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
} DemoCfgT;


typedef struct HdlcStatsT {
  OTX_INT32              nTotalRxFrames;
  OTX_INT32              BadRxFrames;
  OTX_INT32              nLiEventsDuringFrames;
} HdlcStatsT;





/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoConfigureE1T1Devices();
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf0(char *formatStr, ...);
static void DemoLogf1(char *formatStr, ...);
static void DemoLogf2(char *formatStr, ...);
static void DemoLogf3(char *formatStr, ...);
static void DemoLogf4(char *formatStr, ...);
static void DemoLogf5(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);
static void DemoFlushLiEvents(void);
static void DemoLogRecommendedSettings(char *pszSerialNumber, char *pszBrdBuildNumber, char *pszCpldBuildNumbers);

/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
OTX_EVENT   g_hLiNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hLiEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
DemoCfgT    g_demoCfg;
HdlcStatsT  g_demoHdlcStats;
int         g_nTermModeIdxMax = 4;  // Max index of Termination Modes in the OtxT1E1TermModeE enum 

struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[4]; // Used for Gimle-16-PCI-Plus and Thor-8-PCI (not for Thor-2-PCI)
  OTX_HANDLE hLi;
} g_hPhysDevs;


struct {
  OTX_HANDLE hLiIn;
} g_hPipes;


struct {
  OTX_HANDLE hHdlcReceiver;
} g_hLogicDevs;


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;




/*******************************************************************************

  @func

*******************************************************************************/
void DemoPrintSignalSettings(void)
{
  // Signal settings
  char *pszTermMode;

  switch (g_demoCfg.eTermMode) {
  case OTX_T1E1_TERM_120_OHM:
    pszTermMode = "120  ";
    break;
  case OTX_T1E1_TERM_100_OHM:
    pszTermMode = "100  ";
    break;
  case OTX_T1E1_TERM_75_OHM:
    pszTermMode = "75U  ";
    break;
  case OTX_T1E1_TERM_HIGHZ:
    pszTermMode = "HIGHZ";
    break;
  case OTX_T1E1_TERM_75_OHM_BALANCED:
    pszTermMode = "75B  ";
    break;
  default:
    pszTermMode = "???  ";
    break;
  }

  DemoLogf2("RxEQ=%s, TermMode=%s, MonMode=%s, Amp=%s, RxThold=%.4d\n",
    g_demoCfg.bReceiveEqualizer ? "ON " : "OFF",
    pszTermMode,
    g_demoCfg.bMonitorMode ? "ON " : "OFF",
    g_demoCfg.bAmplifier ? "ON " : "OFF",
    g_demoCfg.nRxThreshold
  );

}

/*******************************************************************************

  @func

*******************************************************************************/
void DemoPrintFramerSettings(void)
{
  // Signal settings
  char *pszLineCode = "HDB3";
  char *pszFrameFormat = "CRC4_MULTIFRAME_G706";

  if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
    switch (g_demoCfg.eT1LineCode) {
    case OTX_T1_LC_B8ZS:
      pszLineCode = "B8ZS ";
      break;
    case OTX_T1_LC_AMI:
      pszLineCode = "AMI ";
      break;
    default:
      pszLineCode = "??? ";
      break;
    }
    
    switch (g_demoCfg.eT1FrameFormat) {
    case OTX_T1_FF_F12:
      pszFrameFormat = "F12";
      break;
    case OTX_T1_FF_F4:
      pszFrameFormat = "F4";
      break;
    case OTX_T1_FF_ESF:
      pszFrameFormat = "ESF";
      break;
    case OTX_T1_FF_F72:
      pszFrameFormat = "F72";
      break;
    case OTX_T1_FF_F12_J1:
      pszFrameFormat = "F12_J1";
      break;
    case OTX_T1_FF_F4_J1:
      pszFrameFormat = "F4_J1";
      break;
    case OTX_T1_FF_ESF_J1:
      pszFrameFormat = "ESF_J1";
      break;
    case OTX_T1_FF_F72_J1:
      pszFrameFormat = "F72_J1";
      break;
    default:
      pszFrameFormat = "???";
      break;
    }
  
  } else {  // E1 mode
    switch (g_demoCfg.eE1LineCode) {
    case OTX_E1_LC_HDB3:
      pszLineCode = "HDB3 ";
      break;
    case OTX_E1_LC_AMI:
      pszLineCode = "AMI ";
      break;
    default:
      pszLineCode = "??? ";
      break;
    }
    
    switch (g_demoCfg.eE1FrameFormat) {
    case OTX_E1_FF_CRC4_MULTIFRAME_G706:
      pszFrameFormat = "CRC4_MULTIFRAME_G706";
      break;
    case OTX_E1_FF_CRC4_MULTIFRAME:
      pszFrameFormat = "CRC4_MULTIFRAME     ";
      break;
    case OTX_E1_FF_DOUBLE_FRAME:
      pszFrameFormat = "DOUBLE_FRAME        ";
      break;
    default:
      pszFrameFormat = "???                 ";
      break;
    }
  }

  DemoLogf2("FrameFormat=%s, LineCode=%s\n",
    pszFrameFormat,
    pszLineCode
  );

}



/*******************************************************************************

  @func

  Initialize globals

*******************************************************************************/
void DemoInitalizeConfigStartValues(void)
{
  // Signal settings
  g_demoCfg.bReceiveEqualizer = OTX_FALSE;
  g_demoCfg.eTermMode         = OTX_T1E1_TERM_HIGHZ;
  g_demoCfg.bMonitorMode      = OTX_FALSE;
  g_demoCfg.bAmplifier        = OTX_FALSE;
  g_demoCfg.nRxThreshold      = 910;  // in mV

  // E1 Framer settings
  g_demoCfg.eE1LineCode       = OTX_E1_LC_HDB3;
  g_demoCfg.eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME_G706;

  // T1 Framer settings
  g_demoCfg.eT1LineCode       = OTX_T1_LC_B8ZS;
  g_demoCfg.eT1FrameFormat    = OTX_T1_FF_F72;

  // HDLC settings

}


/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;
                               
  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects  
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);

#if 1
  // Create E1T1 device specific event queue
  result = OtxDrvCreateEvent(&g_hLiNotificationEvent);
  if (OTX_RESULT_CODE(result) == OTX_E_NOTIMPL) {
    g_hLiNotificationEvent = (OTX_EVENT) NULL;    // D16V86 apps cannot use event objects  
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hLiNotificationEvent, &g_hLiEventQueue); DemoCheckRc(g_hLiEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hLiEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hLiEventQueue, result);
#else // Use one global event queue
  g_hLiNotificationEvent = g_hNotificationEvent;
  g_hLiEventQueue = g_hEventQueue;
#endif

}



/*******************************************************************************

  @func
    Open the Physical Devices on the OTX board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; // Only used for Thor-8-PCI
  OTX_UINT32 nDeviceId;
  OTX_UINT32 nSubsystemId;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
    DemoLogf2("Opening Thor-2-PCMCIA board\n");
    break;
  case OTX_DEVICE_THOR_PCI:
    DemoLogf2("Opening Thor-2-PCI board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_PLUS:
    DemoLogf2("Opening Thor-2-PCI-Plus board\n");
    break;
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
    DemoLogf2("Opening Thor-2-PCI-Express board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf2("Opening Thor-8-PCI board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf2("Opening Thor-8-PCI Plus board\n");
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    DemoLogf2("Opening Thor-4-PCI Plus board\n");
    break;
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    DemoLogf2("Opening Sleipnir board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    DemoLogf2("Opening Gimle-16-PCI Plus board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCIE:
    DemoLogf2("Opening Gimle-16-PCIe board\n");
    break;
  default:
    DemoLogf0("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_DEVICE_ID, &nDeviceId); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SUBSYSTEM_ID, &nSubsystemId); DemoCheckRc(g_hPhysDevs.hThorController0, result);
  DemoLogf3("Board Device ID: 0x%.4X, SubsystemId: 0x%.4X\n", nDeviceId, nSubsystemId);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCMCIA:
  case OTX_DEVICE_THOR_PCI:
  case OTX_DEVICE_THOR_2_PCI_PLUS:
  case OTX_DEVICE_THOR_2_PCI_EXPRESS:
  case OTX_DEVICE_SLEIPNIR_1_PCI:
    // On the Thor-2-PCI and Thor-2-PCMCIA boards the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    i = g_demoEnv.nTestLi;
    DemoLogf3("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hLiEventQueue, &g_hPhysDevs.hLi); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    break;

  case OTX_DEVICE_THOR_8_PCI:
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    // On the Thor-8-PCI board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf3("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }

    // Get handles to T1/E1 line interface devices 
    i = g_demoEnv.nTestLi;
    nQuadE1T1 = i/4;
    DemoLogf3("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hLiEventQueue, &g_hPhysDevs.hLi); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
  case OTX_DEVICE_GIMLE_16_PCIE:
    // On the Gimle-16-PCI board the E1T1 devices are hosted by a quad E1T1 device. Open those four devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
      DemoLogf3("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }

    // Get handles to T1/E1 line interface devices 
    i = g_demoEnv.nTestLi;
    nQuadE1T1 = i/4;
    DemoLogf3("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hLiEventQueue, &g_hPhysDevs.hLi); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    break;
  case OTX_DEVICE_THOR_4_PCI_PLUS:
    // On the Thor-4-PCI board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf3("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    }

    // Get handles to T1/E1 line interface devices 
    i = g_demoEnv.nTestLi;
    nQuadE1T1 = i/4;
    DemoLogf3("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hLiEventQueue, &g_hPhysDevs.hLi); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    break;
  default:
    DemoLogf0("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  
}



/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoCfg.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoCfg.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoCfg.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoCfg.eE1FrameFormat;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_TRUE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775
  pE1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = g_demoCfg.bReceiveEqualizer;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;    
  pT1CfgData->eTxLineCode                = g_demoCfg.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoCfg.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoCfg.eT1FrameFormat;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_TRUE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_TRUE;
  pT1CfgData->nLosSensitivity            = 10;  // 10 = LOS detection after 176 consecutive zeros which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLosRecovery               = 21;  // 21 = LOS recovery after 22 ones in the LosRecovery interval which fulfills G.775, Telcordia (Bellcore) and AT&T
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = g_demoCfg.bReceiveEqualizer;
  pT1CfgData->bfClearChannels            = 0x000000; // 0x000000 = No channels are marked as clear channels
}






/*******************************************************************************

  @func
    Configure E1/T1 Devices

*******************************************************************************/
void DemoConfigureE1T1Devices()
{
  OTX_RESULT result;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Set termination for the T1/E1 Front End
  result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi, g_demoCfg.eTermMode); DemoCheckRc(g_hPhysDevs.hLi, result);
  
  if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
    // T1 mode
    result = OtxT1SetConfiguration(g_hPhysDevs.hLi, &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi, result);
  } else {
    // E1 mode
    result = OtxE1SetConfiguration(g_hPhysDevs.hLi, &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi, result);
  }
  result = OtxT1E1SetLiMode(g_hPhysDevs.hLi, g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi, result);
  
  // Tri-state the transmit pins of the Line interface (so it does not disturb the receive pins in High-Z mode)
  result = OtxT1E1XmitPinControl(g_hPhysDevs.hLi, OTX_FALSE); DemoCheckRc(g_hPhysDevs.hLi, result);

  // Set receive threshold
  result = OtxDrvSetAttributeValueUint32(g_hPhysDevs.hLi, OTX_ATTR_T1E1_RX_INPUT_THRESHOLD, g_demoCfg.nRxThreshold); DemoCheckRc(g_hPhysDevs.hLi, result);

  // Set Monitor (-20dB) mode
  result = OtxT1E1SetMonitorMode(g_hPhysDevs.hLi, g_demoCfg.bMonitorMode); DemoCheckRc(g_hPhysDevs.hLi, result);
  
  // Turning front end Amplifier ON/OFF      
  result = OtxDrvSetAttributeValueBool(g_hPhysDevs.hLi, OTX_ATTR_T1E1_RX_AMPLIFIER, g_demoCfg.bAmplifier); DemoCheckRc(g_hPhysDevs.hLi, result);
  
}


/*******************************************************************************

  @func

*******************************************************************************/
OTX_INT32 DemoCheckForRedAlarm()
{
  OTX_INT32 nRedAlarmCount = 0;
  OTX_RESULT result;
  OTX_INT32 i;
#if 0
  OtxT1E1Layer1StatusExS sLayer1Status;

  for (i = 0; i< 10; i++) {
    OsaThreadSleep(20);
    result = OtxT1E1GetLayer1StatusEx(g_hPhysDevs.hLi, &sLayer1Status); DemoCheckRc(g_hPhysDevs.hLi, result);
    if (sLayer1Status.bRedAlarm) {
      nRedAlarmCount++;
    }    
  }
#else  // The device is disabled during the test so we can't use OtxT1E1GetLayer1StatusEx()
  OTX_UINT32 readVal;

  for (i = 0; i< 5; i++) {
    OsaThreadSleep(10);
    result = OtxDrvReadReg8(g_hPhysDevs.hLi, FA_FRS0_REG, &readVal); DemoCheckRc(g_hPhysDevs.hLi, result);
    if (readVal & FRS0_LOS) {
      nRedAlarmCount++;
    }
  }
#endif

  return nRedAlarmCount;
  
}

/*******************************************************************************

  @func

*******************************************************************************/
OTX_INT32 DemoCheckForFraming()
{
  OTX_INT32 i;
  OTX_INT32 nFramingErrorCount = 0;
  OTX_RESULT result;
#if 0
  OtxT1E1Layer1StatusExS sLayer1Status;

  if (g_demoEnv.nFramerTestTime < 200) g_demoEnv.nFramerTestTime = 200;

  for (i = 0; i < g_demoEnv.nFramerTestTime/200; i++) {

    OsaThreadSleep(200);

    result = OtxT1E1GetLayer1StatusEx(g_hPhysDevs.hLi, &sLayer1Status); DemoCheckRc(g_hPhysDevs.hLi, result);
    if (sLayer1Status.bFrameError || sLayer1Status.bMultiFrameError) {
      nFramingErrorCount++;
    } 
    
  }
#else
  OTX_UINT32 readVal;

  if (g_demoEnv.nFramerTestTime < 100) g_demoEnv.nFramerTestTime = 100; // We must run through the loop at least once

  for (i = 0; i < g_demoEnv.nFramerTestTime/100; i++) {

    OsaThreadSleep(100);

    result = OtxDrvReadReg8(g_hPhysDevs.hLi, FA_FRS0_REG, &readVal); DemoCheckRc(g_hPhysDevs.hLi, result);
    //if ((readVal & FRS0_LFA) || (readVal & FRS0_LMFA)) 
    if (readVal & FRS0_LFA) 
    {
      nFramingErrorCount++;
    }

    if ( (g_demoCfg.eT1FrameFormat == OTX_T1_FF_F72) || (g_demoCfg.eT1FrameFormat == OTX_T1_FF_F12) ) {        
      if (readVal & FRS0_LMFA) {
        nFramingErrorCount++;
      }
    }

  }
#endif

  return nFramingErrorCount;
  
}


/*******************************************************************************

  @func

*******************************************************************************/
void DemoConnectAndConfigureHdlcDevices()
{
  OTX_HANDLE hHostDevice;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 ts = g_demoEnv.nTs;

  memset(&g_demoHdlcStats, 0, sizeof(HdlcStatsT));

  DemoFlushLiEvents();

  result = OtxDrvEnable(g_hPhysDevs.hLi, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi, result);


  // Create Hdlc Receiver logical devices hosted in the physical Line Interface devices
  hHostDevice = g_hPhysDevs.hLi;
  result = OtxDrvCreateLogicalDevice(hHostDevice, OTX_LDEVICE_HDLC_RECEIVER, g_hLiEventQueue, 0, &g_hLogicDevs.hHdlcReceiver); DemoCheckRc(hHostDevice, result);

  // Create an incoming Pipe representing the external T1/E1
  result = OtxDrvCreatePipe(64, &g_hPipes.hLiIn); DemoCheckRc(g_hPipes.hLiIn, result);
  
  DemoLogf3("Connecting HDLC Receiver to TS%d\n", ts);

  memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
  sqTimeSlotMask[ts] = 0xFF;
  
  result = OtxDrvConnectPipe(g_hPipes.hLiIn, OTX_PIPE_OUTPUT, g_hPhysDevs.hLi, 1, sqTimeSlotMask, 32); DemoCheckRc(g_hPipes.hLiIn, result);
  
  // Connect HDLC Receiver to the external T1/E1 Pipes
  result = OtxDrvConnectLogicalDevice(g_hPipes.hLiIn, -1, g_hLogicDevs.hHdlcReceiver, -1); DemoCheckRc(g_hPipes.hLiIn, result);

  // Initialize the Hdlc Receiver
  result = OtxHdlcReceiverInit(
    g_hLogicDevs.hHdlcReceiver, // Handle to logical HDLC Filter Device
    OTX_HDLC_FFT_FLAGS,          // Fill character(s) to be send between HDLC frames
    2,                           // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
    OTX_HDLC_CRC16,              // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
    OTX_FALSE,                   // If OTX_TRUE then all bits are inverted when transmitted and before received
    OTX_TRUE,                    // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
    OTX_HDLC_TR_NA,              // HDLC Transmission Rate for V.110 and X.30 modes
    OTX_FALSE);                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
  DemoCheckRc(g_hLogicDevs.hHdlcReceiver, result);
  
  if (g_demoEnv.bSs7FisuFiltering) {
    result = OtxHdlcSS7SetFilter(g_hLogicDevs.hHdlcReceiver, OTX_HDLC_SS7_FILTER_FISU); DemoCheckRc(g_hLogicDevs.hHdlcReceiver, result);
  }

  // Enable Logical HDLC Devices 
  result = OtxDrvEnable(g_hLogicDevs.hHdlcReceiver, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hLogicDevs.hHdlcReceiver, result);


}


/*******************************************************************************

  @func

*******************************************************************************/
void DemoDisconnectHdlcDevices()
{
  OTX_RESULT result;

  result = OtxDrvDisable(g_hPhysDevs.hLi, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi, result);

  // Disconnect HDLC Receiver from the external T1/E1 Pipes
  result = OtxDrvDisconnectLogicalDevice(g_hPipes.hLiIn, -1, g_hLogicDevs.hHdlcReceiver, -1); DemoCheckRc(g_hPipes.hLiIn, result);

  result = OtxDrvDisconnectPipe(g_hPipes.hLiIn, OTX_PIPE_OUTPUT); DemoCheckRc(g_hPipes.hLiIn, result);

  // Close the pipe
  result = OtxDrvClose(g_hPipes.hLiIn); DemoCheckRc(g_hPipes.hLiIn, result);

  // Close Logical HDLC Devices 
  result = OtxDrvClose(g_hLogicDevs.hHdlcReceiver); DemoCheckRc(g_hLogicDevs.hHdlcReceiver, result);

}


/*******************************************************************************

  @func

*******************************************************************************/
OTX_INT32 DemoCheckForBadHdlcFrames()
{
  OTX_INT32 i;

  if (g_demoEnv.nHdlcTestTime < 10) g_demoEnv.nHdlcTestTime = 10; // We must run through the loop at least once

  for (i = 0; i < g_demoEnv.nHdlcTestTime/10; i++) {
    OsaThreadSleep(10);

    if (g_demoHdlcStats.BadRxFrames) {
      return g_demoHdlcStats.BadRxFrames;
    }

  }

  return g_demoHdlcStats.BadRxFrames;

}



/*******************************************************************************

  @func

  Returns OTX_TRUE when a good configuration is found.
  Returns OTX_FALSE when no good configuration was found.

  @devnote

  This implementation relies heavily on the existing enums (and their enum value)
  for line codes and frame types. It was implemented for the following enums:

  enum OtxE1LineCodeE {
    OTX_E1_LC_AMI  = 0, 
    OTX_E1_LC_HDB3 = 1  
  };

  enum OtxT1LineCodeE {
    OTX_T1_LC_AMI   = 0,   
    OTX_T1_LC_B8ZS  = 1    
  };

  enum OtxE1FrameFormatE {
    OTX_E1_FF_DOUBLE_FRAME         = 0,   
    OTX_E1_FF_CRC4_MULTIFRAME      = 1,   
    OTX_E1_FF_CRC4_MULTIFRAME_G706 = 2    
  };

  enum OtxT1FrameFormatE {
    OTX_T1_FF_F12    = 0,  
    OTX_T1_FF_F4     = 1,  
    OTX_T1_FF_ESF    = 2,  
    OTX_T1_FF_F72    = 3,  
    OTX_T1_FF_F12_J1 = 4,  
    OTX_T1_FF_F4_J1  = 5,  
    OTX_T1_FF_ESF_J1 = 6,  
    OTX_T1_FF_F72_J1 = 7,  
  };

  enum OtxT1E1TermModeE {
    OTX_T1E1_TERM_120_OHM,        
    OTX_T1E1_TERM_100_OHM,        
    OTX_T1E1_TERM_75_OHM,         
    OTX_T1E1_TERM_HIGHZ,          
    OTX_T1E1_TERM_75_OHM_BALANCED   (only available in OTX driver 1.0-P150 and later)
  };


*******************************************************************************/
OTX_BOOL DemoAutoDetectLineConfig()
{
  int nHdlcTries;
  int a, b, c, d, e;
  int k,l;

  a = 0;
  b = 1;
  c = 0;
  d = 0;
  e = 910;


TRY_SIGNAL_SETTINGS:

  for (; a <= 1; a++) { // bReceiveEqualizer
    for (; b <= g_nTermModeIdxMax; b++) { // eTermMode, skip 120 ohm since that is the same hardware setting as 100 ohm (to save time)
      for (; c <= 1; c++) { // bMonitorMode
        for (; d <= 1; d++) { // bAmplifier
          if (c && a) { // If we have both MonitorMode and ReceiveEqualizer, then the driver will pick the RxThreshold. No need to loop through the RxThreshold settings in this case.
            continue;
          }

          for (; e > 0; ) {  // RxThreshold
            
            g_demoCfg.bReceiveEqualizer = (OTX_BOOL) a;
            g_demoCfg.eTermMode         = (OtxT1E1TermModeE) b;
            g_demoCfg.bMonitorMode      = (OTX_BOOL) c;
            g_demoCfg.bAmplifier        = (OTX_BOOL) d;
            g_demoCfg.nRxThreshold      = e;  // in mV
            
            DemoLogf2("Trying SignalCfg: "); DemoPrintSignalSettings();
            DemoConfigureE1T1Devices();

            if (e > 350) {
              e = e - 300;
            } else if 
              (e > 100) { 
              e = e - 60;
            } else { 
              e = e - 30;
            }

            if (DemoCheckForRedAlarm() == 0) {
              DemoLogf2("Signal settings are OK\n"); 
              goto TRY_FRAMER_SETTINGS;
            }

            
          }
          e = 910;
          
        }
        d = 0;
      }
      c = 0;
    }
    b = 1;
  }
  
  DemoLogf1("Tried all Signal settings to no avail.\n");
  return OTX_FALSE;


TRY_FRAMER_SETTINGS:

  k = 1;

  if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_E1) {
    l = 2;  // 3 elements in OtxE1FrameFormatE
  } else {
    l = 3;  // 8 elements in OtxT1FrameFormatE, however, we don't check J1 
  }

TRY_FRAMER_SETTINGS_AGAIN:

  for (; k >= 0; k--) {  // E1 mode: starts with OTX_E1_LC_HDB3, and then try OTX_E1_LC_AMI; see OtxT1LineCodeE for T1 mode 
    for (; l >= 0;) {    // E1 mode: starts with OTX_E1_FF_CRC4_MULTIFRAME_G706, then OTX_E1_FF_CRC4_MULTIFRAME, and then OTX_E1_FF_DOUBLE_FRAME last; see OtxT1FrameFormatE for T1 mode
      
      if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_E1) {
        g_demoCfg.eE1LineCode       = (OtxE1LineCodeE) k;
        g_demoCfg.eE1FrameFormat    = (OtxE1FrameFormatE) l;
      } else {
        g_demoCfg.eT1LineCode       = (OtxT1LineCodeE) k;
        g_demoCfg.eT1FrameFormat    = (OtxT1FrameFormatE) l;
      }
      
      DemoLogf2("Trying FramerCfg: "); DemoPrintFramerSettings();
      DemoConfigureE1T1Devices();

      l--;

      if (DemoCheckForFraming() == 0) {
        DemoLogf2("Framer settings are OK\n"); 
        goto TRY_HDLC_SETTINGS;
      }
      
    }
    
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_E1) {
      l = 2;
    } else {
      l = 3;
    }
  }

  DemoLogf2("Framer errors occurred. Trying another signal setting.\n"); 
  goto TRY_SIGNAL_SETTINGS;


TRY_HDLC_SETTINGS:

  // No HDLC settings at this time. Although we could add a loop for for inverted/non-inverted HDLC bits

  DemoConnectAndConfigureHdlcDevices();

  for (nHdlcTries = 0; nHdlcTries < 2; nHdlcTries++) {
    DemoLogf2("Checking HDLC frames...\n");
    if (DemoCheckForBadHdlcFrames() == 0) {
      DemoLogf2("HDLC frames check ok (bad=%d total=%d, liEvents=%d)\n", g_demoHdlcStats.BadRxFrames, g_demoHdlcStats.nTotalRxFrames, g_demoHdlcStats.nLiEventsDuringFrames);

      if (g_demoHdlcStats.nLiEventsDuringFrames) {
        DemoLogf2("Li events were received. Trying HDLC check again with the same settings\n");
        memset(&g_demoHdlcStats, 0, sizeof(HdlcStatsT));
        continue;
      }

      if (g_demoHdlcStats.nTotalRxFrames == 0) {
        DemoLogf1("*** Warning: No HDLC frames were received during the HDLC test.\n");
        DemoLogf1("*** Does TS%d really carry HDLC traffic?\n", g_demoEnv.nTs);
      }

      return OTX_TRUE; // We found a good configuration
    }
  }

  DemoDisconnectHdlcDevices(); // Get ready to try a different setting

  DemoLogf2("HDLC errors occurred. Trying another framer setting.\n"); 
  goto TRY_FRAMER_SETTINGS_AGAIN;

}




/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  nQuadE1T1;

  // Enable Board
  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // If board type is Thor-8-PCI then enable Quad E1T1 device
  if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI || g_demoEnv.nBoardType == OTX_DEVICE_THOR_8_PCI_PLUS) {
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf3("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  } else if (g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCI_PLUS || g_demoEnv.nBoardType == OTX_DEVICE_GIMLE_16_PCIE) {
    // Enable Quad E1T1 devices
    for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
      DemoLogf3("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  } else if (g_demoEnv.nBoardType == OTX_DEVICE_THOR_4_PCI_PLUS) {
    // Enable Quad E1T1 device
    for (nQuadE1T1 = 0; nQuadE1T1 < 1; nQuadE1T1++) {
      DemoLogf3("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
      result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
    }
  }

}



/*******************************************************************************
  @func
    Flush Events
*******************************************************************************/
void DemoFlushLiEvents()
{
#define DEMO_EVENT_DATA_BUF_SIZE 256
  OTX_RESULT result;

  do {
    OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
    OTX_INT32 coEventsFetched;
    
    result = OtxDrvGetEventDataEx(g_hLiEventQueue, sqEventData, DEMO_EVENT_DATA_BUF_SIZE, &coEventsFetched); DemoCheckRc(g_hLiEventQueue, result);
    
  } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  
}




/*******************************************************************************
  @func
    Gets Driver Event (if any)
*******************************************************************************/
void *DemoGetEvents(
  void *lpParameter   // thread data
)
{
#define DEMO_EVENT_DATA_BUF_SIZE 256
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    DemoLogf5("Global Event signalled\n");
    do {
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;
      
      result = OtxDrvGetEventDataEx(g_hEventQueue, sqEventData, DEMO_EVENT_DATA_BUF_SIZE, &coEventsFetched); DemoCheckRc(g_hEventQueue, result);
      
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)
        DemoLogf5("OtxDrvGetEventDataEx() fetched %d events\n", coEventsFetched);
        
        assert(coEventsFetched <= coEventsFetched);
        
        for (i = 0; i < coEventsFetched; i++) {
          DemoLogf0("Unhandled event from device handle 0x%X, deviceType=0x%X\n", sqEventData[i].m_hDevice, sqEventData[i].m_eDeviceType);
        }
        
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
    
  }
  
  return 0;
}


/*******************************************************************************
  @func
    Gets Li related Events (if any)
*******************************************************************************/
void *DemoGetLiEvents(
  void *lpParameter   // thread data
)
{
#define DEMO_EVENT_DATA_BUF_SIZE 256
  OTX_RESULT result;

  while (1) {
    OtxDrvWaitForSingleEvent(g_hLiNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf5("Event signalled\n");
    do {
      OtxEventDataS sqEventData[DEMO_EVENT_DATA_BUF_SIZE];
      OTX_INT32 coEventsFetched;
      OTX_INT32 i;
      
      result = OtxDrvGetEventDataEx(g_hLiEventQueue, sqEventData, DEMO_EVENT_DATA_BUF_SIZE, &coEventsFetched); DemoCheckRc(g_hEventQueue, result);
      
      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event (one at least)
        //DemoLogf5("OtxDrvGetEventDataEx() fetched %d events\n", coEventsFetched);
        
        assert(coEventsFetched <= coEventsFetched);
        
        for (i = 0; i < coEventsFetched; i++) {
          // Print the event data.
          // In a real application branch based on the data
          DemoPrintEvent(&sqEventData[i]);
        }
        
      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
    
  }

  return 0;
  
}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  char msgStr[256];
    
  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_LI_T1E1:
    // Event Li

    g_demoHdlcStats.nLiEventsDuringFrames++;

    if (g_demoEnv.nVerboseLevel >= 4) {
      
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (!strncmp(msgStr, "Invalid", strlen("Invalid"))) {
        DemoLogf4("E1/T1[%d] Code is %d\n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
      if (OTX_SUCCESS(otxRc)) {
        if (pEventData->m_nParam) {
          DemoLogf4("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
        } else {
          DemoLogf4("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
        }
      } else {
        DemoLogf4("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;

  case OTX_LDEVICE_HDLC_RECEIVER:
    // Event from HLDC Receiver 
    if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
#define CB_BUF_SIZE 8192
      OTX_UINT8  buf[CB_BUF_SIZE];
      OTX_UINT8  *pBuf;
			OtxHdlcFrameInfoS *pFrame;
      OTX_UINT32 coFrames;
      OTX_UINT32 j;
      
      otxRc = OtxHdlcReadDataEx(pEventData->m_hDevice, buf, CB_BUF_SIZE, DEMO_MAX_MULTIPLE_FRAME_READ, OTX_FALSE, &coFrames);
      while (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
      
        pFrame = (OtxHdlcFrameInfoS *) buf;

        for (j = 0; j < coFrames; j++) {

          g_demoHdlcStats.nTotalRxFrames++;

					pBuf = pFrame->sqData;

          if (pFrame->nFrameEndCode) {
            g_demoHdlcStats.BadRxFrames++;
          }
              
          if (g_demoEnv.nVerboseLevel >= 5) {
            DemoLogf5("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, pFrame->cbOctets, pFrame->nFrameEndCode, pFrame->nFrameSeq);
            for (i = 0; i < pFrame->cbOctets && i < CB_BUF_SIZE; i++) {
              DemoLogf5("%.2X ", pBuf[i]);
            }
            DemoLogf5("\n");
          }  

          // Step pointer to next frame (if any)
					pFrame = OTX_STRUCT_ADD_BYTE_OFFSET(OtxHdlcFrameInfoS, pFrame, pFrame->nOffsetNext);
        }

        // Read next frame (if any)
        otxRc = OtxHdlcReadDataEx(pEventData->m_hDevice, buf, CB_BUF_SIZE, DEMO_MAX_MULTIPLE_FRAME_READ, OTX_FALSE, &coFrames);

      } 
      if (OTX_RESULT_CODE(otxRc) != OTX_S_FALSE) {
        DemoLogf0("HdlcReceiver[%d] OtxHdlcReadDataEx() was unsuccessful.\n", pEventData->m_nSourceId);
        DemoCheckRc(pEventData->m_hDevice, otxRc);
      }
      
      
    } else {
      otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf0("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      } else {
        DemoLogf0("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
      }
    }
    break;
    
  default:
    // Unhandled event
    DemoLogf0("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}

/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->nBoardType        = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->nTestLi           = 0;
  envP->nTs               = 0; // 0=use default
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->bSs7FisuFiltering = OTX_FALSE;
  envP->nVerboseLevel     = 4; // The higher the level the more verbose output
  envP->nHdlcTestTime     = DEFAULT_HDLC_TEST_TIME; // in milliseconds
  envP->nFramerTestTime   = DEFAULT_FRAMER_TEST_TIME; // in milliseconds


  strcpy(envP->logFileName, envP->cwd);
  strcat(envP->logFileName, DEMO_DEFAULT_LOG_FILE);

  strcpy(envP->cfgFileName, envP->cwd);
  strcat(envP->cfgFileName, DEMO_DEFAULT_CFG_FILE);

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {          // Log file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'C') {   // Cfg file
          strcpy(envP->logFileName, argv[i] + 2);
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'V') {   // Verbose level
          envP->nVerboseLevel = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == '2') {
            envP->nBoardType = OTX_DEVICE_THOR_PCI;
          } else if (*(argv[i] + 2) == '8') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
          } else if (*(argv[i] + 2) == 'T') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_PLUS;
          } else if (*(argv[i] + 2) == 'E') {
            envP->nBoardType  = OTX_DEVICE_THOR_2_PCI_EXPRESS;
          } else if (*(argv[i] + 2) == 'G') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCI_PLUS;
          } else if (*(argv[i] + 2) == '9') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCIE;
          } else if (*(argv[i] + 2) == 'P') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
          } else if (*(argv[i] + 2) == '4') {
            envP->nBoardType  = OTX_DEVICE_THOR_4_PCI_PLUS;
          } else if (*(argv[i] + 2) == 'M') {
            envP->nBoardType  = OTX_DEVICE_THOR_PCMCIA;
          } else if (*(argv[i] + 2) == 'S') {
            envP->nBoardType  = OTX_DEVICE_SLEIPNIR_1_PCI;
          } else {
            DemoLogf0("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf0("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
          /*
        } else if (*(argv[i]+1) == 'S') {   // SS7 FISU filtering
          envP->bSs7FisuFiltering = OTX_TRUE;
          */
        } else if (*(argv[i]+1) == 'N') {   // HDLC TX
          envP->nTs = (OTX_UINT16)strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'I') {   // Test Li number
          envP->nTestLi = (OTX_UINT16)strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'H') {   // Hdlc Test time
          envP->nHdlcTestTime = (OTX_UINT16)strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'F') {   // Framer Test time
          envP->nFramerTestTime = (OTX_UINT16)strtoul(argv[i]+2, NULL, 0);
        } else {
          DemoLogf0("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf0("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }


}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf0("\nCommand line switches:\n");
  DemoLogf0(" /B<boardNo>    : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf0(" /T[2|T|E|8|G|9|P|M|S] : Board Type used for the demo, where\n");
  DemoLogf0("                    2  = Use Thor-2-PCI Board\n");
  DemoLogf0("                    T  = Use Thor-2-PCI-Plus Board\n");
  DemoLogf0("                    E  = Use Thor-2-PCI-Express Board\n");
  DemoLogf0("                    8  = Use Thor-8-PCI Board\n");
  DemoLogf0("                    G  = Use Gimle-16-PCI-Plus Board\n");
  DemoLogf0("                    9  = Use Gimle-16-PCIe Board\n");
  DemoLogf0("                    P  = Use Thor-8-PCI-Plus Board\n");
  DemoLogf0("                    4  = Use Thor-4-PCI-Plus Board\n");
  DemoLogf0("                    M  = Use Thor-2-PCMCIA Board\n");
  DemoLogf0("                    S  = Use Sleipnir Board\n");
  DemoLogf0(" /L<fileName>   : Log file name (default is \"%s\")\n", DEMO_DEFAULT_LOG_FILE);
  DemoLogf0(" /C<fileName>   : Cfg output file name (default is \"%s\")\n", DEMO_DEFAULT_CFG_FILE);
  //DemoLogf0(" /S             : SS7 FISU filtering active\n");
  DemoLogf0(" /ME            : E1 Li mode\n");
  DemoLogf0(" /MT            : T1 Li mode\n");
  DemoLogf0(" /V<num>        : Verbose level (0=quiet, 5=very verbose, default is %d)\n", g_demoEnv.nVerboseLevel);
  DemoLogf0(" /N<ts>         : Timeslot to connect the HDLC receiver to\n");
  DemoLogf0(" /I<li>         : Li number to run the auto config on\n");
  DemoLogf0(" /H<time>       : HDLC test time in milliseconds. Default is %d ms.\n", DEFAULT_HDLC_TEST_TIME);
  DemoLogf0(" /F<time>       : Framer test time in milliseconds. Default is %d ms.\n", DEFAULT_FRAMER_TEST_TIME);
  DemoLogf0(" /Z<str>        : Command string (overrides keyboard input)\n");
  DemoLogf0(" /Q             : Exit without prompting on errors\n");
}


/*******************************************************************************
  @func
*******************************************************************************/
static void DemoLogf0(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}

/*******************************************************************************
  @func
*******************************************************************************/
static void DemoLogf1(char *formatStr, ...)
{
  va_list argList;

  if (g_demoEnv.nVerboseLevel >= 1) {

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  }

}

/*******************************************************************************
  @func
*******************************************************************************/
static void DemoLogf2(char *formatStr, ...)
{
  va_list argList;

  if (g_demoEnv.nVerboseLevel >= 2) {

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  }

}

/*******************************************************************************
  @func
*******************************************************************************/
static void DemoLogf3(char *formatStr, ...)
{
  va_list argList;

  if (g_demoEnv.nVerboseLevel >= 3) {

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  }

}

/*******************************************************************************
  @func
*******************************************************************************/
static void DemoLogf4(char *formatStr, ...)
{
  va_list argList;

  if (g_demoEnv.nVerboseLevel >= 4) {

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  }

}

/*******************************************************************************
  @func
*******************************************************************************/
static void DemoLogf5(char *formatStr, ...)
{
  va_list argList;

  if (g_demoEnv.nVerboseLevel >= 5) {

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  }

}


/*******************************************************************************

  @func

  Displays a message on the screen 

  @comm

*******************************************************************************/
static void DemoPrintf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);

  va_end(argList);

}

static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf0("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(OTX_DRIVER_DEVICE_HANDLE, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf0("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf0("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}




/*******************************************************************************
*******************************************************************************/
void DemoAtExit(void)
{
  fclose(g_pDemoLogFile);
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_RESULT result;  // OTX Standard Return Code
  char szSerialNumber[128];
  char szBrdBuildNumber[128];
  char szCpldBuildNumbers[128];

  OsaConsoleInit();

  DemoLogf1("T8pLineAutoDetect, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf2("E1T1 Device Example program.\n");
  DemoLogf2("For Thor-2-PCI-Plus, Thor-8-PCI-Plus, Gimle-16-PCI-Plus, and Thor-2-PCMCIA-PRO\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    DemoPrintf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }
  atexit(DemoAtExit);

  DemoInitalizeConfigStartValues();

  if (!g_demoEnv.nTs) {
    // TS not specified. Use a default TS depending on the T1/E1 mode.
    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      g_demoEnv.nTs = 24; 
    } else {
      // E1 mode
      g_demoEnv.nTs = 16; 
    }
  }

  DemoLogf2("Log file is   %s\n", g_demoEnv.logFileName);
  DemoLogf2("Board number  %d\n", g_demoEnv.boardNo);
  DemoLogf2("Li            %d\n", g_demoEnv.nTestLi);
  DemoLogf2("HDLC Timeslot %d\n", g_demoEnv.nTs);


  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf3("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }

  // Connect to the Driver
  DemoLogf3("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf0("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf0("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Query driver revision
  {
    char szVersionInfo[255];
    OtxVersionInfo sVersionInfo;
    OTX_UINT32 cbAttribValue;

    result = OtxDrvGetAttributeValue(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_REV_NUM, &sVersionInfo, sizeof(OtxVersionInfo), &cbAttribValue); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

    result = OtxDrvGetAttributeValueSz(OTX_DRIVER_DEVICE_HANDLE, OTX_ATTR_DRV_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf3("Driver version %s  (Major=%d, Minor=%d, RelPrel=%d, nPrelLevel=%d)\n", szVersionInfo, sVersionInfo.nMjRev, sVersionInfo.nMnRev, sVersionInfo.bRelStatus, sVersionInfo.nPrelLevel);

    if ((sVersionInfo.nMjRev == 1) && (sVersionInfo.nMnRev == 0) && (sVersionInfo.bRelStatus == 0) && sVersionInfo.nPrelLevel < 150) {
      // In driver versions earlier than 1.0-P150 the OTX_T1E1_TERM_75_OHM_BALANCED mode did not exist.
      // Adjust for this fact.
      DemoLogf3("TermMode OTX_T1E1_TERM_75_OHM_BALANCED does not exist in this driver version. Adjusting.\n");
      g_nTermModeIdxMax = 3;
    }
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf4("Creating event queue...\n");
  DemoCreateEventQueue();

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf4("Opening physical devices...\n");
  DemoOpenPhysicalDevices();

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoEnablePhysicalDevices();

  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, (OtxBrdClockSourceE) ((int)OTX_CLOCK_SOURCE_LOCAL_0 + g_demoEnv.nTestLi)); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Read Serial, Board Build, CPLD Build Numbers from the board
  {
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, szSerialNumber, 128); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf3("Board Serial Number: %s\n", szSerialNumber);

    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, szBrdBuildNumber, 128); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf3("Board Build Number: %s\n", szBrdBuildNumber);

    if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
      result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, szCpldBuildNumbers, 128); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      DemoLogf3("CPLD Build:\n%s", szCpldBuildNumbers);
    } else {
      szCpldBuildNumbers[0] = '\0';
    }
  }

  OsaThreadStart(DemoGetEvents);

  OsaThreadStart(DemoGetLiEvents);

  // Try different line settings until the Line interface does not report any status messages
  if (DemoAutoDetectLineConfig()) {
    DemoLogf1("\nAuto configuration successful.\n");
    DemoLogf1("The recommended signal settings are:\n");
    DemoPrintSignalSettings();
    DemoLogf1("The recommended framer settings are:\n");
    DemoPrintFramerSettings();
    DemoLogRecommendedSettings(szSerialNumber, szBrdBuildNumber, szCpldBuildNumbers);
    
    OtxDrvReleaseLib();
    OsaConsoleExit();
    
    return 0;
  } else {
    DemoLogf1("Auto configuration failed.\n");
    
    OtxDrvReleaseLib();
    OsaConsoleExit();
    
    return 1;
  }

}

/*******************************************************************************

  @func

*******************************************************************************/
void DemoLogRecommendedSettings(
  char *pszSerialNumber,
  char *pszBrdBuildNumber,
  char *pszCpldBuildNumbers
)
{
  FILE *fCfg;
  char *pszTermMode;
  char *pszLineCode;
  char *pszFrameFormat;
  char *tmp;

  fCfg = fopen(g_demoEnv.cfgFileName, "at");
  if (!fCfg) {
    DemoLogf0("Error: Could write the recommended settings to file \"%s\"\n", g_demoEnv.cfgFileName);
    return;
  }

  fprintf(fCfg, "-------------------------------------------------------------------\n\n");

  fprintf(fCfg, "Board number        %d\n", g_demoEnv.boardNo);
  fprintf(fCfg, "Li                  %d\n\n", g_demoEnv.nTestLi);

  fprintf(fCfg, "Serial number       \"%s\"\n", pszSerialNumber);
  fprintf(fCfg, "Board build number  \"%s\"\n", pszBrdBuildNumber);
  fprintf(fCfg, "Cpld build numbers  \"");
  tmp = strtok(pszCpldBuildNumbers,  "\n");
  while (tmp) {
    fprintf(fCfg, "%s", tmp);
    tmp = strtok(NULL, "\n");
    if (tmp) fprintf(fCfg, ", ");
  }
  fprintf(fCfg, "\"\n\n");

  fprintf(fCfg, "Recommended configuration settings are:\n\n");

  switch (g_demoCfg.eTermMode) {
  case OTX_T1E1_TERM_120_OHM:
    pszTermMode = "120  ";
    break;
  case OTX_T1E1_TERM_100_OHM:
    pszTermMode = "100  ";
    break;
  case OTX_T1E1_TERM_75_OHM:
    pszTermMode = "75U  ";
    break;
  case OTX_T1E1_TERM_HIGHZ:
    pszTermMode = "HIGHZ";
    break;
  case OTX_T1E1_TERM_75_OHM_BALANCED:
    pszTermMode = "75B  ";
    break;
  default:
    pszTermMode = "???  ";
    break;
  }

  fprintf(fCfg, "Receive Equalizer       = %s\n",
    g_demoCfg.bReceiveEqualizer ? "ON " : "OFF"
  );
  fprintf(fCfg, "Termination Mode        = %s\n",
    pszTermMode
  );
  fprintf(fCfg, "Monitor Mode            = %s\n",
    g_demoCfg.bMonitorMode ? "ON " : "OFF"
  );
  fprintf(fCfg, "Front End Amplifier     = %s\n",
    g_demoCfg.bAmplifier ? "ON " : "OFF"
  );
  fprintf(fCfg, "Receive Input Threshold = %.4d mV\n",
    g_demoCfg.nRxThreshold
  );


  // E1 mode
  switch (g_demoCfg.eE1LineCode) {
  case OTX_E1_LC_HDB3:
    pszLineCode = "HDB3 ";
    break;
  case OTX_E1_LC_AMI:
    pszLineCode = "AMI ";
    break;
  default:
    pszLineCode = "??? ";
    break;
  }
  
  switch (g_demoCfg.eE1FrameFormat) {
  case OTX_E1_FF_CRC4_MULTIFRAME_G706:
    pszFrameFormat = "CRC4_MULTIFRAME_G706";
    break;
  case OTX_E1_FF_CRC4_MULTIFRAME:
    pszFrameFormat = "CRC4_MULTIFRAME     ";
    break;
  case OTX_E1_FF_DOUBLE_FRAME:
    pszFrameFormat = "DOUBLE_FRAME        ";
    break;
  default:
    pszFrameFormat = "???                 ";
    break;
  }

  // T1 mode
  switch (g_demoCfg.eT1LineCode) {
  case OTX_T1_LC_B8ZS:
    pszLineCode = "B8ZS ";
    break;
  case OTX_T1_LC_AMI:
    pszLineCode = "AMI ";
    break;
  default:
    pszLineCode = "??? ";
    break;
  }
    
  switch (g_demoCfg.eT1FrameFormat) {
  case OTX_T1_FF_F12:
    pszFrameFormat = "F12";
    break;
  case OTX_T1_FF_F4:
    pszFrameFormat = "F4";
    break;
  case OTX_T1_FF_ESF:
    pszFrameFormat = "ESF";
    break;
  case OTX_T1_FF_F72:
    pszFrameFormat = "F72";
    break;
  case OTX_T1_FF_F12_J1:
    pszFrameFormat = "F12_J1";
    break;
  case OTX_T1_FF_F4_J1:
    pszFrameFormat = "F4_J1";
    break;
  case OTX_T1_FF_ESF_J1:
    pszFrameFormat = "ESF_J1";
    break;
  case OTX_T1_FF_F72_J1:
    pszFrameFormat = "F72_J1";
    break;
  default:
    pszFrameFormat = "???                 ";
    break;
  }


  fprintf(fCfg, "Frame Format            = %s\n",
    pszFrameFormat
  );
  fprintf(fCfg, "Line Code               = %s\n",
    pszLineCode
  );

  fprintf(fCfg, "\n");
  fprintf(fCfg, "-------------------------------------------------------------------\n");

  fclose(fCfg);

  DemoLogf1("The recommended settings are written to file \"%s\"\n", g_demoEnv.cfgFileName);


}

